from __future__ import annotations

from typing import List, Optional

from pytest import raises
from typer import Typer

from terracomp_typer._build import _evaluate_type_hints


def test__evaluate_type_hints() -> None:
    def func(value: str | None) -> list[str]:
        assert value == "foo"
        return [value]

    func = _evaluate_type_hints(func)

    assert func.__annotations__ == {"value": Optional[str], "return": List[str]}

    # Ensure that Typer can understand the type hints on the function now.
    app = Typer()
    app.command()(func)
    with raises(SystemExit) as excinfo:
        app(args=["foo"])
    assert excinfo.value.code == 0



"""
Implement dependency injection resolution on functions.
"""

from __future__ import annotations

from dataclasses import dataclass
from functools import wraps
from inspect import signature
from typing import Any, Callable, Container, Dict, List, Set, Type, TypeVar, cast, get_type_hints

from nr.stream import Supplier

T = TypeVar("T")


class DependencyInjectionError(Exception):
    pass


class DependencyInjector:
    """
    Arranges for dependency injection based on function signatures.
    """

    @dataclass
    class _Declared:
        pass

    @dataclass
    class _Instance:
        value: Any

    @dataclass
    class _Supplier:
        value: Callable[[], Any]

    _mapping: Dict[Type[Any], _Declared | _Instance | _Supplier]
    _delayed: Set[Type[Any]]

    def __init__(self, *objects: object) -> None:
        self._mapping = {}
        self._delayed = set()
        for obj in objects:
            if type(obj) in self._mapping:
                raise TypeError(
                    "cannot populate dependency provider with multiple instances of the same type "
                    f"({type(obj).__name__})"
                )
            self._mapping[type(obj)] = self._Instance(obj)

    def declare(self, type_: Type[T]) -> None:
        """
        Declare that at a later point, an instance or supplier for the given *type_* will be registered.
        When a function is processed by #bound() and it references types that are only declared, the resolution
        of the type will occur at call time of the function.
        """

        if type_ in self._mapping:
            raise TypeError(f"slot for type {type_.__name__} is already allocated in dependency injector")
        self._mapping[type_] = self._Declared()

    def register_supplier(self, type_: Type[T], supplier: Callable[..., T]) -> None:
        """
        Register a supplier for a given type. Any arguments expected by the *supplier* are resolved
        using the same dependency injection mechanism.
        """

        existing = self._mapping.get(type_)
        if existing is not None and not isinstance(existing, self._Declared):
            raise TypeError(f"slot for type {type_.__name__} is already allocated in dependency injector")
        self._mapping[type_] = self._Supplier(Supplier.once(Supplier.of_callable(self.bind(supplier))).get)

    def get_dependency_for_type(self, type_: Type[T]) -> T:
        """
        Resolve the value of a given type.
        """

        if not isinstance(type_, type):
            raise DependencyInjectionError(f"cannot provide dependency for non-type: {type_!r}")

        if type_ == DependencyInjector:
            return cast(T, self)

        if type_ in self._mapping:
            value = self._mapping[type_]
            if isinstance(value, self._Declared):
                raise DependencyInjectionError(
                    f"unable to provide a dependency for type {type_.__name__}, but the type was declared"
                )
            elif isinstance(value, self._Supplier):
                return cast(T, value.value())
            else:
                return cast(T, value.value)

        raise DependencyInjectionError(f"unable to provide a dependency for type {type_.__name__}")

    def bind(
        self,
        func: Callable[..., Any],
        allow_unresolved: bool = False,
        ignore: Container[str] = (),
    ) -> Callable[..., Any]:
        """
        Bind dependencies requested through the annotations of *func*.

        :param func: The function to bind dependencies to.
        :param allow_unresolved: Allow unresolved parameters and keep them in the function signature.
        :param ignore: Ignore parameters with these names.
        """

        undefined = object()

        sig = signature(func)
        annotations = get_type_hints(func)
        return_annotation = annotations.pop("return", undefined)

        # Inject the dependencies.
        remaining = {}
        bindings = {}
        delayed = set()
        for key, value in list(annotations.items()):
            # TODO(NiklasRosenstein): We should probably somehow derive a new instance of the dependency
            #       injector that has the current one as a parent delegate instead of globally declaring
            #       the bindings defined here.
            default = sig.parameters[key].default
            if value == DependencyInjector and isinstance(default, _DelayedBinding):
                for type_ in default.types:
                    self.declare(type_)

            if key in ignore:
                remaining[key] = value
            elif isinstance(self._mapping.get(value), (self._Declared, self._Supplier)):
                delayed.add(key)
            else:
                try:
                    bindings[key] = self.get_dependency_for_type(value)
                except DependencyInjectionError:
                    if not allow_unresolved:
                        raise
                    remaining[key] = value

        if not allow_unresolved and remaining:
            raise DependencyInjectionError(...)  # TODO

        if return_annotation is not undefined:
            remaining["return"] = return_annotation

        if not bindings and not delayed:
            return func

        @wraps(func)
        def _wrapper(*args: Any, **kwargs: Any) -> Any:
            for key in delayed:
                kwargs[key] = self.get_dependency_for_type(annotations[key])
            return func(*args, **kwargs, **bindings)

        _wrapper.__annotations__ = remaining
        _wrapper.__signature__ = sig.replace(parameters=[v for k, v in sig.parameters.items() if k not in bindings and k not in delayed])  # type: ignore[attr-defined]  # noqa: E501
        return _wrapper


@dataclass
class _DelayedBinding:
    types: List[Type[Any]]


def DelayedBinding(*types: Type[Any]) -> Any:
    """
    This function should be used as the default value on a parameter that expects a #DependencyInjector, indicating
    that the function will supply the injector with additional dependencies of the given types.
    """

    return _DelayedBinding(list(types))



from __future__ import annotations

import asyncio
from functools import wraps
from importlib import import_module
from inspect import Signature, iscoroutinefunction, signature
from pathlib import Path
from pkgutil import iter_modules
from typing import Any, Callable, Coroutine, Mapping, Sequence

from typeapi import get_annotations
from typer import Exit, Typer

from ._injector import DependencyInjector


def build_app_from_module(
    module_name: str,
    name: str | None = None,
    typer_options: Mapping[str, Any] | None = None,
    dependencies: DependencyInjector | Sequence[Any] = (),
    event_loop: asyncio.AbstractEventLoop | None = None,
) -> Typer:
    """
    Looks at the module given with *module_name* and adds subcommand groups or commands to the Typer *app* based on
    the contents. Packages with an `__init__.py` will create a subcommand group, where the docstring of that module
    is the help of the group. Python modules that don't start with an underscore will create a command in the current
    subcommand group.

    :param module_name: The module to create a #Typer application for. The module must be a package with submodules.
    :param name: Override the name of the root #Typer application.
    :param typer_options: Keyword arguments to pass to every #Typer creation.
    :param dependencies: A sequence of objects with unique types that are injected into commands based on their
        function signature.
    """

    if event_loop is None:
        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)

    if not isinstance(dependencies, DependencyInjector):
        dependencies = DependencyInjector(*dependencies)

    module = import_module(module_name)
    assert module.__file__, f"module {module_name!r} has no __file__"
    assert Path(module.__file__).stem == "__init__", f"expected a package for {module_name!r}"

    name = name or module_name.rpartition(".")[-1]
    app = Typer(name=name, help=module.__doc__, **(typer_options or {}))

    if hasattr(module, "callback"):
        app.callback()(_prepare_typer_func(module.callback, dependencies, event_loop))

    for submodule_info in iter_modules(module.__path__, prefix=module_name + "."):
        subcommand_name = submodule_info.name.rpartition(".")[-1].replace("_", "-")
        module_spec = submodule_info.module_finder.find_spec(submodule_info.name, module.__path__)  # type: ignore
        submodule = import_module(submodule_info.name)
        assert module_spec is not None, f"unable to find module spec of {submodule_info.name!r}"
        assert module_spec.origin is not None, f"module spec of {submodule_info.name!r} has no origin"

        if Path(module_spec.origin).stem == "__init__":
            sub_app = build_app_from_module(
                submodule_info.name,
                typer_options=typer_options,
                dependencies=dependencies,
                event_loop=event_loop,
            )
            app.add_typer(sub_app)

        elif not submodule_info.name.startswith("_"):
            func = _prepare_typer_func(submodule.main, dependencies, event_loop)
            app.command(
                name=subcommand_name,
                help=func.__doc__ or submodule.__doc__,
            )(func)

    return app


def _prepare_typer_func(
    func: Callable[..., int | None], dependencies: DependencyInjector, event_loop: asyncio.AbstractEventLoop
) -> Callable[..., None]:
    func = _evaluate_type_hints(func)
    if iscoroutinefunction(func):
        func = _deasyncify(func, event_loop)
    func = dependencies.bind(func, allow_unresolved=True)
    func = _raise_non_zero_exit_code_as_exit(func)
    return func


def _evaluate_type_hints(func: Callable[..., Any]) -> Callable[..., Any]:
    annotations = get_annotations(func)
    sig = signature(func)
    sig = sig.replace(
        parameters=[p.replace(annotation=annotations.get(p.name, p.empty)) for p in sig.parameters.values()],
        return_annotation=annotations.get("return", Signature.empty),
    )
    func.__signature__ = sig  # type: ignore[attr-defined]
    func.__annotations__ = annotations
    return func


def _raise_non_zero_exit_code_as_exit(func: Callable[..., int | None]) -> Callable[..., None]:
    """
    Wraps a function to raise a #Exit exception at the end.
    """

    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        result = func(*args, **kwargs)
        if isinstance(result, int):
            raise Exit(code=result)
        elif result is not None:
            raise RuntimeError("expected None or integer return value from command function")

    return wrapper


def _deasyncify(
    func: Callable[..., Coroutine[Any, Any, Any]], event_loop: asyncio.AbstractEventLoop
) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        return event_loop.run_until_complete(func(*args, **kwargs))

    return wrapper



__version__ = "0.0.5"

from ._build import build_app_from_module
from ._injector import DelayedBinding, DependencyInjectionError, DependencyInjector

__all__ = [
    "DependencyInjectionError",
    "DependencyInjector",
    "DelayedBinding",
    "build_app_from_module",
]
